import warnings
from typing import Type


class W:
    def __init__(self, msg: str, warning_cls: Type[Warning]):
        self.msg = msg
        self.warning_cls = warning_cls

    def __call__(self, *args, **kwargs):
        warnings.warn(self.msg.format(*args, **kwargs), self.warning_cls, stacklevel=3)


class E:
    def __init__(self, msg: str, exception_cls: Type[Exception]):
        self.msg = msg
        self.exception_cls = exception_cls

    def __call__(self, *args, **kwargs):
        raise self.exception_cls(
            self.msg.format(*args, **kwargs), self.warning_cls, stacklevel=3
        )


class Warnings:
    _W_FOR_TEST = W("Foo {} wow {bar}!", RuntimeWarning)
    W0 = W("{} has been deprecated. Please use {} instead.", DeprecationWarning)
    W1 = W("{} has been deprecated.", DeprecationWarning)


class Errors:
    E0 = E(
        "Requires {package}. Please install it with `pip install {package}`.",
        ImportError,
    )



from functools import lru_cache
from pathlib import Path
from typing import Any, Iterable, List, Optional, Tuple, cast

import numpy as np
import spacy
import torch
import torch.nn.functional as F
import transformers.modeling_bert
from spacy.gold import GoldParse
from spacy.language import Language
from spacy.pipeline import Pipe
from spacy.tokens import Doc
from transformers import BertConfig

from camphr.torch_utils import TorchPipe, add_loss_to_docs

from .model import TRANSFORMERS_MODEL
from .tokenizer import TRANSFORMERS_TOKENIZER, TrfTokenizer
from .utils import SerializationMixinForTrfTask, get_last_hidden_state_from_docs

MASKEDLM_PREDICTION = "maskedlm_prediction"
MASKEDLM_LABEL = "maskedlm_label"
BERT = "bert"


def get_maskedlm_labels(docs: List[Doc]) -> torch.Tensor:
    return docs[0].user_data[MASKEDLM_LABEL]


def set_maskedlm_labels(docs: List[Doc], labels: torch.Tensor):
    docs[0].user_data[MASKEDLM_LABEL] = labels


class BertOnlyMLMHead(transformers.modeling_bert.BertOnlyMLMHead):
    def __init__(self, config):
        super().__init__(config)
        self.config = config


class PIPES:
    bert_for_maskedlm_preprocessor = "bert_for_maskedlm_preprocessor"
    bert_for_maskedlm = "bert_for_maskedlm"


@spacy.component(PIPES.bert_for_maskedlm_preprocessor)
class BertForMaskedLMPreprocessor(Pipe):
    def __init__(self, vocab, model=True, **cfg):
        self.vocab = vocab
        self.cfg = cfg
        self.model = model
        self._exclude_ids: Optional[np.ndarray] = None

    @property
    def p_mask(self) -> float:
        return self.cfg.get(
            "p_mask", 0.15 * 0.8
        )  # ref: https://arxiv.org/pdf/1905.05583.pdf

    @property
    def p_replace(self) -> float:
        return self.cfg.get(
            "p_replace", 0.15 * 0.1
        )  # ref: https://arxiv.org/pdf/1905.05583.pdf

    @property
    def p_dist(self) -> List[float]:
        p_remain = 1 - self.p_mask - self.p_replace
        return [p_remain, self.p_mask, self.p_replace]

    @property  # type: ignore
    @lru_cache()
    def exclude_ids(self) -> np.ndarray:
        return np.array(self.model.all_special_ids)

    def __call__(self, doc: Doc):
        return doc

    def pipe(self, docs: Iterable[Doc], *args, **kwargs):
        return docs

    def update(self, docs: List[Doc], *args, **kwargs):  # type: ignore
        self.require_model()
        inputs = TrfTokenizer.get_transformers_input(docs)
        assert inputs is not None
        input_ids = inputs.input_ids
        set_maskedlm_labels(docs, input_ids)

        masked_input_ids = input_ids.clone()
        target = torch.from_numpy(~np.isin(masked_input_ids, self.exclude_ids))
        mask_idx, replace_idx = self._choice_labels(masked_input_ids, target)

        masked_input_ids[mask_idx] = self.model.mask_token_id
        masked_input_ids[replace_idx] = torch.randint(
            low=max(self.model.all_special_ids),
            high=self.model.vocab_size,
            size=(cast(int, replace_idx.sum().cpu().item()),),
        )
        inputs.input_ids = masked_input_ids

    def _choice_labels(
        self, wordpieces: torch.Tensor, target: torch.Tensor
    ) -> Tuple[torch.Tensor, torch.Tensor]:
        labels = torch.zeros_like(wordpieces, device=wordpieces.device)
        label = np.random.choice(3, target.sum().cpu().item(), p=self.p_dist)
        label = torch.from_numpy(label).to(device=wordpieces.device)
        labels[target] = label
        return labels == 1, labels == 2  # mask index, replace index

    def to_disk(self, path: Path, *args, **kwargs):
        # This component has nothing to be saved.
        # You should call `add_maskedlm` pipe after restoring.
        pass


@spacy.component(PIPES.bert_for_maskedlm)
class BertForMaskedLM(SerializationMixinForTrfTask, TorchPipe):
    model_cls = BertOnlyMLMHead
    trf_config_cls = BertConfig

    def predict(self, docs: Iterable[Doc]):
        self.require_model()
        self.model.eval()
        x = get_last_hidden_state_from_docs(docs)
        with torch.no_grad():
            preds: torch.Tensor = self.model(x)
        assert len(preds.shape) == 3  # (b,l,h)
        return preds

    def set_annotations(self, docs: Iterable[Doc], preds: Any):
        for doc, pred in zip(docs, preds):
            doc.user_data[MASKEDLM_PREDICTION] = torch.max(pred)

    def update(  # type: ignore
        self, docs: List[Doc], golds: Iterable[GoldParse], **kwargs
    ):
        self.require_model()
        self.model.train()
        x = get_last_hidden_state_from_docs(docs)
        preds: torch.Tensor = self.model(x)
        preds = preds.view(-1, self.model.config.vocab_size)
        targets = (
            get_maskedlm_labels(docs).view(-1)[: len(preds)].to(device=self.device)
        )
        loss = F.cross_entropy(preds, targets, ignore_index=-1)
        add_loss_to_docs(docs, loss)


def add_maskedlm_pipe(nlp: Language):
    """Add maskedlm pipe to nlp"""
    wp = nlp.get_pipe(TRANSFORMERS_TOKENIZER)
    tokenizer = wp.model
    preprocessor = BertForMaskedLMPreprocessor(nlp.vocab, tokenizer)
    nlp.add_pipe(preprocessor, before=TRANSFORMERS_MODEL)
    bert = nlp.get_pipe(TRANSFORMERS_MODEL)
    config = bert.model.config
    model = BertOnlyMLMHead(config)
    pipe = BertForMaskedLM(nlp.vocab, model)
    nlp.add_pipe(pipe)


def remove_maskedlm_pipe(nlp: Language):
    """Remove maskedlm pipe from nlp"""
    targets = [PIPES.bert_for_maskedlm, PIPES.bert_for_maskedlm_preprocessor]
    for k in targets:
        if k in nlp.pipe_names:
            nlp.remove_pipe(k)



"""Defines utility functions, classes, mixins for transformers pipelines."""
import dataclasses
import pickle
from contextlib import contextmanager
from pathlib import Path
from typing import (
    Any,
    Callable,
    Dict,
    Generic,
    Iterable,
    Iterator,
    List,
    Sequence,
    Sized,
    Type,
    TypeVar,
    cast,
)

import torch
import torch.nn as nn
import transformers
from spacy.gold import GoldParse
from spacy.language import Language
from spacy.tokens import Doc
from spacy.vocab import Vocab
from tokenizations import get_alignments
from typing_extensions import Literal, Protocol

from camphr.pipelines.utils import UserHooksMixin
from camphr.torch_utils import TensorWrapper, set_grad

from .auto import get_trf_config_cls, get_trf_name


class ATTRS:
    """Attribute names for spacy.Underscore"""

    tokens = "transformers_tokens"
    token_ids = "transformers_token_ids"
    cleaned_tokens = "transformers_cleaned_tokens"
    special_tokens = "transformers_special_tokens"
    align = "transformers_align"
    last_hidden_state = "transformers_last_hidden_state"
    batch_inputs = "transformers_batch_inputs"


def _get_transformers_align(doc: Doc) -> List[List[int]]:
    """Get tokens alignment from spacy tokens to transformers tokens"""
    trf_tokens = doc._.get(ATTRS.cleaned_tokens)
    return get_alignments([token.text for token in doc], trf_tokens)[0]


def _set_extensions():
    for attr in [
        ATTRS.tokens,
        ATTRS.token_ids,
        ATTRS.cleaned_tokens,
        ATTRS.last_hidden_state,
        ATTRS.batch_inputs,
    ]:
        Doc.set_extension(attr, default=None)
    Doc.set_extension(ATTRS.align, getter=_get_transformers_align)


_set_extensions()


TRF_CONFIG = "trf_config"
VOCAB_SIZE = "vocab_size"
TRF_NAME = "trf_name"
CONVERT_LABEL = "convert_label"
LABEL_WEIGHTS = "label_weights"
LABELS = "labels"


def get_last_hidden_state_from_docs(docs: Iterable[Doc]) -> torch.Tensor:
    """Get transformers text embedding from docs.
    
    Useful for downstream task pipelines.
    `last_hidden_state` is set in `camphr.pipelines.transformers.TrfModel.set_annotations`.
    """
    # assumed that the batch tensor of all docs is stored in the extension.
    x: TensorWrapper = next(iter(docs))._.get(ATTRS.last_hidden_state)
    if x is None:
        raise ValueError("docs don't have `last_hidden_state`.")
    return x.batch_tensor


def get_dropout(config: transformers.PretrainedConfig) -> float:
    """Get dropout rate from config"""
    if isinstance(config, transformers.BertConfig):
        return config.hidden_dropout_prob
    if hasattr(config, "dropout"):
        return config.dropout
    return 0.1


class TrfModelForTaskBase(nn.Module):
    def __init__(self, config: transformers.PretrainedConfig):
        super().__init__()
        self.config = config


class _TrfSavePathGetter:
    def _trf_path(self, path: Path) -> Path:
        return path / self.cfg[TRF_NAME]  # type: ignore


class SerializationMixinForTrfTask(_TrfSavePathGetter):
    """Mixin for transoformers pipeline

    Bounds: `TorchPipe`
    """

    _MODEL_PTH = "model.pth"
    _CFG_PKL = "cfg.pkl"
    model_cls: Type[TrfModelForTaskBase]

    def to_disk(self, path: Path, exclude=tuple(), **kwargs):
        path.mkdir(exist_ok=True)
        model: TrfModelForTaskBase = self.model
        config_save_path = self._trf_path(path)
        config_save_path.mkdir(exist_ok=True)
        model.config.save_pretrained(str(config_save_path))
        torch.save(model.state_dict(), str(path / self._MODEL_PTH))

        with (path / self._CFG_PKL).open("wb") as f:
            cfg = {k: v for k, v in self.cfg.items() if k not in exclude}
            pickle.dump(cfg, f)

    def from_disk(self, path: Path, exclude=tuple(), **kwargs):
        with (path / self._CFG_PKL).open("rb") as f:
            self.cfg = pickle.load(f)

        config = get_trf_config_cls(self.cfg[TRF_NAME]).from_pretrained(
            str(self._trf_path(path))
        )
        model = self.model_cls(config)
        model.load_state_dict(
            torch.load(
                str(path / self._MODEL_PTH), map_location=self.device  # type: ignore
            )
        )
        model.eval()
        self.model = model
        return self


class FromNLPMixinForTrfTask:
    """Mixin for transformers task pipeline.

    Bounds: `TorchPipe`
    """

    @classmethod
    def from_pretrained(cls, vocab: Vocab, trf_name_or_path: str, **cfg):
        """Load pretrained model."""
        name = get_trf_name(trf_name_or_path)
        return cls(  # type: ignore
            vocab,
            model=cls.Model(trf_name_or_path, **cfg),  # type: ignore
            trf_name=name,
            **cfg
        )

    @classmethod
    def from_nlp(cls, nlp: Language, **cfg):
        if cfg.get("trf_name_or_path"):
            return cls.from_pretrained(nlp.vocab, **cfg)
        return cls(nlp.vocab)  # type: ignore


class PipeProtocol(Protocol):
    """For mypy. See https://mypy.readthedocs.io/en/latest/more_types.html#mixin-classes"""

    cfg: Dict[str, Any]


class LabelsMixin:
    """Mixin for pipes which has labels.


    Bounds: `Pipe`
    Optional Bounds:
        - `UserHooksMixin`: to use `def convert_label`
    """

    @property
    def labels(self: PipeProtocol) -> List[str]:
        # The reason to return empty list is that `spacy.Language` calls `Pipe.labels` when restoring,
        # before the `labels` is not set to this Pipe.
        # For the same reason cache cannot be used.
        return self.cfg.get(LABELS, [])

    @property
    def label2id(self: PipeProtocol) -> Dict[str, int]:
        # `lru_cache` is not adopted because mypy doesn't support `Decorated property`
        # https://github.com/python/mypy/issues/1362
        if not hasattr(self, "_label2id"):
            self._label2id = {v: i for i, v in enumerate(self.labels)}  # type: ignore
        return self._label2id  # type: ignore

    @property
    def label_weights(self: PipeProtocol) -> torch.Tensor:
        if not hasattr(self, "_label_weights"):
            weights_map = self.cfg.get(LABEL_WEIGHTS)
            weights = torch.ones(len(self.label2id))  # type: ignore
            if weights_map:
                assert len(weights_map) == len(self.label2id)  # type: ignore
                for k, v in weights_map.items():
                    weights[self.label2id[k]] = v  # type: ignore
            self._label_weights = weights
        return self._label_weights  # type: ignore

    def convert_label(self: UserHooksMixin, label: str) -> str:  # type:ignore
        fn = self.user_hooks.get(CONVERT_LABEL)
        if fn:
            return fn(label)
        return label


class TrfObj(Protocol):
    @classmethod
    def from_pretrained(
        cls, pretrained_model_name_or_path, *inputs, **kwargs
    ) -> "TrfObj":
        ...

    def save_pretrained(self, *args, **kwargs):
        ...


T_TrfObj = TypeVar("T_TrfObj", bound=TrfObj)


class TrfAutoMixin(_TrfSavePathGetter, Generic[T_TrfObj]):
    """Mixin for transformers' `AutoModel` and `AutoTokenizer`

    Required to be combined with `TorchPipe`
    """

    _MODEL_PTH = "model.pth"
    _CFG_PKL = "cfg.pkl"
    _MODEL_CLS_GETTER: Callable[[str], Type[T_TrfObj]]
    model: T_TrfObj

    @classmethod
    def Model(cls, trf_name_or_path: str, **cfg):
        _cls = cls._MODEL_CLS_GETTER(trf_name_or_path)
        return _cls.from_pretrained(trf_name_or_path, **cfg)

    @classmethod
    def from_pretrained(cls, vocab: Vocab, trf_name_or_path: str, **cfg):
        """Load pretrained model."""
        name = get_trf_name(trf_name_or_path)
        model = cls(  # type: ignore
            vocab, model=cls.Model(trf_name_or_path), trf_name=name, **cfg
        )
        return model

    @classmethod
    def from_nlp(cls, nlp: Language, **cfg):
        if cfg.get("trf_name_or_path"):
            return cls.from_pretrained(nlp.vocab, **cfg)
        return cls(nlp.vocab)  # type: ignore

    def to_disk(self, path: Path, exclude=tuple(), **kwargs):
        path.mkdir(exist_ok=True)
        model_path = self._trf_path(path)
        model_path.mkdir(exist_ok=True)
        self.model.save_pretrained(str(model_path))
        with (path / self._CFG_PKL).open("wb") as f:
            pickle.dump(self.cfg, f)

    def from_disk(self, path: Path, exclude=tuple(), **kwargs):
        with (path / self._CFG_PKL).open("rb") as f:
            self.cfg = pickle.load(f)
        self.model = self.Model(str(self._trf_path(path)))
        return self


T = TypeVar("T")


class EstimatorMixin(Generic[T]):
    """Mixin for downstream pipelines.

    This mixin provides functionalities common to `train`, `eval`, and `predict`.
    The followings is assumed:

        1. `train` computes loss under `train` mode.
        2. `eval` computes loss under `eval` mode and sets annotation.
        3. `predict` doesn't compute loss and sets annotation.

    At a minimum, you should implement the below methods that raises `NotImplementedError`.
    See `ner.TrfForNamedEntityRecognition` for example usage.

    Bounds:
        - attributes: `set_annotations`, `model`, `require_model`
        - `model` is subclass of `torch.nn.Module`
    """

    def proc_model(self, docs: Iterable[Doc]) -> T:
        raise NotImplementedError

    def compute_loss(
        self, docs: Sequence[Doc], golds: Sequence[GoldParse], outputs: T
    ) -> None:
        raise NotImplementedError

    def update(self, docs: Sequence[Doc], golds: Sequence[GoldParse], **kwargs) -> None:
        with self.switch("train"):
            outputs = self.proc_model(docs)
            self.compute_loss(docs, golds, outputs)

    def eval(self, docs: List[Doc], golds: List[GoldParse]) -> None:
        with self.switch("eval"):
            outputs = self.proc_model(docs)
            self.compute_loss(docs, golds, outputs)
            self.set_annotations(docs, outputs)  # type: ignore

    def predict(self, docs: List[Doc]) -> T:
        with self.switch("eval"):
            return self.proc_model(docs)

    @contextmanager
    def switch(self, mode: Literal["train", "eval"]):
        self.require_model()  # type: ignore
        if mode == "train":
            self.model.train()  # type: ignore
            grad = True
        else:
            self.model.eval()  # type: ignore
            grad = False
        with set_grad(grad):
            yield


@dataclasses.dataclass
class TransformersInput:
    input_ids: torch.Tensor
    token_type_ids: torch.Tensor
    attention_mask: torch.Tensor
    input_len: torch.LongTensor

    def __iter__(self) -> Iterator["TransformersInput"]:
        for i in range(len(cast(Sized, self.input_ids))):
            _row = {k: getattr(self, k)[i] for k in self.tensor_field_names}
            row = TransformersInput(**_row)
            yield row

    def to(self, *args, **kwargs):
        for k in self.tensor_field_names:
            t = getattr(self, k).to(*args, **kwargs)
            setattr(self, k, t)

    @property
    def tensor_field_names(self) -> List[str]:
        return [
            field.name
            for field in dataclasses.fields(self)
            if field.type is torch.Tensor or field.type is torch.LongTensor
        ]

    @property
    def model_input(self) -> Dict[str, torch.Tensor]:
        output = {k: getattr(self, k) for k in self.tensor_field_names}
        del output["input_len"]
        return output



"""Defines transformers sequence classification pipe"""
import operator
from typing import Any, Dict, Iterable, List, Optional, Sequence, Tuple, cast

import spacy
import torch
import torch.nn as nn
import torch.nn.functional as F
import transformers
from spacy.gold import GoldParse
from spacy.tokens import Doc
from transformers.modeling_utils import SequenceSummary

from camphr.pipelines.utils import UserHooksMixin
from camphr.torch_utils import TorchPipe, add_loss_to_docs, goldcat_to_label

from .auto import get_trf_config_cls
from .utils import (
    ATTRS,
    EstimatorMixin,
    FromNLPMixinForTrfTask,
    LabelsMixin,
    SerializationMixinForTrfTask,
    TrfModelForTaskBase,
    get_last_hidden_state_from_docs,
)

spacy.language.ENABLE_PIPELINE_ANALYSIS = True
NUM_SEQUENCE_LABELS = "num_sequence_labels"
LABELS = "labels"
TOP_LABEL = "top_label"
TOPK_LABELS = "topk_labels"


class TrfSequenceClassifier(TrfModelForTaskBase):
    """Head layer for sequence classification task"""

    def __init__(self, config: transformers.PretrainedConfig):
        super().__init__(config)
        self.config = config
        assert hasattr(config, NUM_SEQUENCE_LABELS)
        self.num_labels = getattr(config, NUM_SEQUENCE_LABELS)
        self.sequence_summary = SequenceSummary(config)
        self.classifier = nn.Linear(config.hidden_size, self.num_labels)

    def forward(self, x: torch.Tensor) -> torch.Tensor:  # type: ignore
        x = self.sequence_summary(x)
        logits = self.classifier(x)

        return logits


TRANSFORMERS_SEQ_CLASSIFIER = "transformers_sequence_classifier"


@spacy.component(
    TRANSFORMERS_SEQ_CLASSIFIER,
    requires=[f"doc._.{ATTRS.last_hidden_state}"],
    assigns=["doc.cats"],
)
class TrfForSequenceClassification(
    EstimatorMixin[torch.Tensor],
    LabelsMixin,
    FromNLPMixinForTrfTask,
    UserHooksMixin,
    SerializationMixinForTrfTask,
    TorchPipe,
):
    """Base class for sequence classification task (e.g. sentiment analysis).

    Requires `TrfModel` before this model in the pipeline.
    """

    model_cls = TrfSequenceClassifier

    @classmethod
    def Model(cls, trf_name_or_path: str, **cfg) -> TrfSequenceClassifier:
        config = get_trf_config_cls(trf_name_or_path).from_pretrained(trf_name_or_path)
        if LABELS in cfg:
            setattr(config, NUM_SEQUENCE_LABELS, len(cfg[LABELS]))
        return TrfSequenceClassifier(config)

    def proc_model(self, docs: Iterable[Doc]) -> torch.Tensor:
        self.require_model()
        x = get_last_hidden_state_from_docs(docs)
        logits = self.model(x)
        assert len(logits.shape) == 2  # (len(docs), num_class)
        return logits

    def set_annotations(self, docs: Iterable[Doc], logits: torch.Tensor):
        probs = torch.softmax(logits, 1)
        for doc, prob in zip(docs, cast(Iterable, probs)):
            doc.cats = self.get_cats_from_prob(prob)

    def get_cats_from_prob(self, prob: torch.Tensor) -> Dict[str, float]:
        assert len(prob.shape) == 1
        return dict(zip(self.labels, prob.tolist()))

    def golds_to_tensor(self, golds: Iterable[GoldParse]) -> torch.Tensor:
        labels = (goldcat_to_label(gold.cats) for gold in golds)
        labels = (self.convert_label(label) for label in labels)
        targets = [self.label2id[label] for label in labels]
        return torch.tensor(targets, device=self.device)

    def compute_loss(
        self, docs: Sequence[Doc], golds: Sequence[GoldParse], outputs: torch.Tensor
    ) -> None:
        self.require_model()
        get_last_hidden_state_from_docs(docs)
        targets = self.golds_to_tensor(golds)
        weight = self.label_weights.to(device=self.device)  # type: ignore

        loss = F.cross_entropy(outputs, targets, weight=weight)
        add_loss_to_docs(docs, loss)


def _top_label(doc: Doc) -> Optional[str]:
    if not doc.cats:
        return None
    return max(doc.cats.items(), key=operator.itemgetter(1))[0]


def _topk_labels(doc: Doc, k: int) -> List[Tuple[str, Any]]:
    if not doc.cats:
        return []
    return sorted(doc.cats.items(), key=operator.itemgetter(1), reverse=True)[:k]


Doc.set_extension(TOP_LABEL, getter=_top_label, force=True)
Doc.set_extension(TOPK_LABELS, method=_topk_labels, force=True)



import json
import logging
import os
import random
from pathlib import Path
from typing import Any, Dict, List, Sequence, Tuple, Union

import hydra
import srsly
from omegaconf import Config
from sklearn.model_selection import train_test_split

from camphr.utils import get_by_dotkey

GoldParsable = Dict[str, Any]
InputData = List[Tuple[str, GoldParsable]]

logger = logging.getLogger(__name__)


def create_data(cfg: Config) -> Tuple[InputData, InputData]:
    data = list(srsly.read_jsonl(Path(cfg.path).expanduser()))
    if cfg.ndata > 0:
        data = random.sample(data, k=cfg.ndata)
    else:
        cfg.ndata = len(data)
    train, val = train_test_split(data, test_size=cfg.val_size)
    srsly.write_jsonl(Path.cwd() / f"train-data.jsonl", train)
    srsly.write_jsonl(Path.cwd() / f"val-data.jsonl", val)
    return train, val


def report_fail(json_serializable_data: Any) -> None:
    fail_path = Path("fail.json").absolute()
    with fail_path.open("w") as f:
        json.dump(json_serializable_data, f, ensure_ascii=False)
        logger.error(f"Error raised. The input data is saved in {str(fail_path)}")


def convert_fullpath_if_path(text: str) -> str:
    path = os.path.expanduser(text)
    try:
        path = hydra.utils.to_absolute_path(path)
    except AttributeError:
        # Not in hydra runtime
        pass
    if os.path.exists(path):
        return path
    return text


def check_nonempty(cfg: Config, fields: Sequence[Union[str, Sequence[str]]]):
    errors = []
    for key in fields:
        if isinstance(key, str) and (not get_by_dotkey(cfg, key)):
            errors.append(f"{key} is required.")
        elif isinstance(key, list) and (not any(get_by_dotkey(cfg, k) for k in key)):
            errors.append(f"Any of {', '.join(key)} is required.")
    if errors:
        raise ValueError("\n".join(errors))



"""Elmo: Deep contextualized word representations (https://allennlp.org/elmo)"""
import json
import shutil
from pathlib import Path
from typing import Generator, Iterable, List

import numpy as np
import spacy
from spacy.pipeline import Pipe
from spacy.tokens import Doc

from camphr.pipelines.allennlp_base import Pathlike
from camphr.pipelines.utils import (
    get_doc_vector_via_tensor,
    get_similarity,
    get_span_vector_via_tensor,
    get_token_vector_via_tensor,
)


@spacy.component("elmo", assigns=["doc.tensor", "doc.vector", "token.vector"])
class Elmo(Pipe):
    WEIGHTS_FILE_NAME = "weights.hdf5"
    OPTIONS_FILE_NAME = "options.json"

    def __init__(self, model=None, **cfg):
        self.model = model
        self.cfg = cfg

    @classmethod
    def from_elmofiles(cls, options_file: Pathlike, weight_file: Pathlike, **cfg):
        """Construct Elmo with elmo files.
        
        Note:
            Elmo files are distributed in https://allennlp.org/elmo
        
        Example:
            >>> elmo = Elmo.from_elmofiles("./options.json", "./weights.hd5")
            >>> nlp = spacy.blank("en")
            >>> nlp.add_pipe(elmo)
        """
        model = cls.Model(options_file, weight_file)
        return cls(model, **cfg)

    @classmethod
    def Model(cls, options_file: Pathlike, weight_file: Pathlike, **cfg):
        from allennlp.commands.elmo import ElmoEmbedder

        return ElmoEmbedder(
            str(Path(options_file).absolute()), str(Path(weight_file).absolute())
        )

    def to_disk(self, path: Pathlike, **cfg):
        path = Path(path)
        path.mkdir(exist_ok=True)
        shutil.copy(
            self.model.elmo_bilm._token_embedder._weight_file,
            str(path / self.WEIGHTS_FILE_NAME),
        )
        with (path / self.OPTIONS_FILE_NAME).open("w") as f:
            json.dump(self.model.elmo_bilm._token_embedder._options, f)

    def from_disk(self, path: Pathlike, **cfg):
        path = Path(path)
        self.model = self.Model(
            path / self.OPTIONS_FILE_NAME, path / self.WEIGHTS_FILE_NAME
        )

    def predict(self, docs: List[Doc]) -> Generator[np.ndarray, None, None]:
        return self.model.embed_sentences([token.text for token in doc] for doc in docs)

    def set_annotations(self, docs: List[Doc], outputs: Iterable[np.ndarray]):
        for doc, vec in zip(docs, outputs):
            assert vec.shape[1] == len(doc), f"vector: {vec.shape}, doc: {len(doc)}"
            doc.tensor = vec[-1]
            doc.user_hooks["vector"] = get_doc_vector_via_tensor
            doc.user_span_hooks["vector"] = get_span_vector_via_tensor
            doc.user_token_hooks["vector"] = get_token_vector_via_tensor
            doc.user_hooks["similarity"] = get_similarity
            doc.user_span_hooks["similarity"] = get_similarity
            doc.user_token_hooks["similarity"] = get_similarity
            doc.user_data["elmo"] = vec



NAME = "NAME"
COLOR = "NAME/COLOR"
COLOR_OTHER = "NAME/COLOR/COLOR_OTHER"
NATURE_COLOR = "NAME/COLOR/NATURE_COLOR"
DISEASE = "NAME/DISEASE"
ANIMAL_DISEASE = "NAME/DISEASE/ANIMAL_DISEASE"
DISEASE_OTHER = "NAME/DISEASE/DISEASE_OTHER"
EVENT = "NAME/EVENT"
EVENT_OTHER = "NAME/EVENT/EVENT_OTHER"
INCIDENT = "NAME/EVENT/INCIDENT"
INCIDENT_OTHER = "NAME/EVENT/INCIDENT/INCIDENT_OTHER"
WAR = "NAME/EVENT/INCIDENT/WAR"
NATURAL_PHENOMENON = "NAME/EVENT/NATURAL_PHENOMENON"
EARTHQUAKE = "NAME/EVENT/NATURAL_PHENOMENON/EARTHQUAKE"
NATURAL_DISASTER = "NAME/EVENT/NATURAL_PHENOMENON/NATURAL_DISASTER"
NATURAL_PHENOMENON_OTHER = "NAME/EVENT/NATURAL_PHENOMENON/NATURAL_PHENOMENON_OTHER"
OCCASION = "NAME/EVENT/OCCASION"
CONFERENCE = "NAME/EVENT/OCCASION/CONFERENCE"
GAME = "NAME/EVENT/OCCASION/GAME"
OCCASION_OTHER = "NAME/EVENT/OCCASION/OCCASION_OTHER"
RELIGIOUS_FESTIVAL = "NAME/EVENT/OCCASION/RELIGIOUS_FESTIVAL"
FACILITY = "NAME/FACILITY"
ARCHAEOLOGICAL_PLACE = "NAME/FACILITY/ARCHAEOLOGICAL_PLACE"
ARCHAEOLOGICAL_PLACE_OTHER = (
    "NAME/FACILITY/ARCHAEOLOGICAL_PLACE/ARCHAEOLOGICAL_PLACE_OTHER"
)
TUMULUS = "NAME/FACILITY/ARCHAEOLOGICAL_PLACE/TUMULUS"
FACILITY_OTHER = "NAME/FACILITY/FACILITY_OTHER"
FACILITY_PART = "NAME/FACILITY/FACILITY_PART"
GOE = "NAME/FACILITY/GOE"
AIRPORT = "NAME/FACILITY/GOE/AIRPORT"
AMUSEMENT_PARK = "NAME/FACILITY/GOE/AMUSEMENT_PARK"
CAR_STOP = "NAME/FACILITY/GOE/CAR_STOP"
GOE_OTHER = "NAME/FACILITY/GOE/GOE_OTHER"
MARKET = "NAME/FACILITY/GOE/MARKET"
MUSEUM = "NAME/FACILITY/GOE/MUSEUM"
PARK = "NAME/FACILITY/GOE/PARK"
PORT = "NAME/FACILITY/GOE/PORT"
PUBLIC_INSTITUTION = "NAME/FACILITY/GOE/PUBLIC_INSTITUTION"
RESEARCH_INSTITUTE = "NAME/FACILITY/GOE/RESEARCH_INSTITUTE"
SCHOOL = "NAME/FACILITY/GOE/SCHOOL"
SPORTS_FACILITY = "NAME/FACILITY/GOE/SPORTS_FACILITY"
STATION = "NAME/FACILITY/GOE/STATION"
THEATER = "NAME/FACILITY/GOE/THEATER"
WORSHIP_PLACE = "NAME/FACILITY/GOE/WORSHIP_PLACE"
ZOO = "NAME/FACILITY/GOE/ZOO"
LINE = "NAME/FACILITY/LINE"
BRIDGE = "NAME/FACILITY/LINE/BRIDGE"
CANAL = "NAME/FACILITY/LINE/CANAL"
LINE_OTHER = "NAME/FACILITY/LINE/LINE_OTHER"
RAILROAD = "NAME/FACILITY/LINE/RAILROAD"
ROAD = "NAME/FACILITY/LINE/ROAD"
TUNNEL = "NAME/FACILITY/LINE/TUNNEL"
WATER_ROUTE = "NAME/FACILITY/LINE/WATER_ROUTE"
GOD = "NAME/GOD"
LOCATION = "NAME/LOCATION"
ADDRESS = "NAME/LOCATION/ADDRESS"
ADDRESS_OTHER = "NAME/LOCATION/ADDRESS/ADDRESS_OTHER"
EMAIL = "NAME/LOCATION/ADDRESS/EMAIL"
PHONE_NUMBER = "NAME/LOCATION/ADDRESS/PHONE_NUMBER"
POSTAL_ADDRESS = "NAME/LOCATION/ADDRESS/POSTAL_ADDRESS"
URL = "NAME/LOCATION/ADDRESS/URL"
ASTRAL_BODY = "NAME/LOCATION/ASTRAL_BODY"
ASTRAL_BODY_OTHER = "NAME/LOCATION/ASTRAL_BODY/ASTRAL_BODY_OTHER"
CONSTELLATION = "NAME/LOCATION/ASTRAL_BODY/CONSTELLATION"
PLANET = "NAME/LOCATION/ASTRAL_BODY/PLANET"
STAR = "NAME/LOCATION/ASTRAL_BODY/STAR"
GEOLOGICAL_REGION = "NAME/LOCATION/GEOLOGICAL_REGION"
BAY = "NAME/LOCATION/GEOLOGICAL_REGION/BAY"
GEOLOGICAL_REGION_OTHER = "NAME/LOCATION/GEOLOGICAL_REGION/GEOLOGICAL_REGION_OTHER"
ISLAND = "NAME/LOCATION/GEOLOGICAL_REGION/ISLAND"
LAKE = "NAME/LOCATION/GEOLOGICAL_REGION/LAKE"
MOUNTAIN = "NAME/LOCATION/GEOLOGICAL_REGION/MOUNTAIN"
RIVER = "NAME/LOCATION/GEOLOGICAL_REGION/RIVER"
SEA = "NAME/LOCATION/GEOLOGICAL_REGION/SEA"
GPE = "NAME/LOCATION/GPE"
CITY = "NAME/LOCATION/GPE/CITY"
COUNTRY = "NAME/LOCATION/GPE/COUNTRY"
COUNTY = "NAME/LOCATION/GPE/COUNTY"
GPE_OTHER = "NAME/LOCATION/GPE/GPE_OTHER"
PROVINCE = "NAME/LOCATION/GPE/PROVINCE"
LOCATION_OTHER = "NAME/LOCATION/LOCATION_OTHER"
REGION = "NAME/LOCATION/REGION"
CONTINENTAL_REGION = "NAME/LOCATION/REGION/CONTINENTAL_REGION"
DOMESTIC_REGION = "NAME/LOCATION/REGION/DOMESTIC_REGION"
REGION_OTHER = "NAME/LOCATION/REGION/REGION_OTHER"
SPA = "NAME/LOCATION/SPA"
NAME_OTHER = "NAME/NAME_OTHER"
NATURAL_OBJECT = "NAME/NATURAL_OBJECT"
COMPOUND = "NAME/NATURAL_OBJECT/COMPOUND"
ELEMENT = "NAME/NATURAL_OBJECT/ELEMENT"
LIVING_THING = "NAME/NATURAL_OBJECT/LIVING_THING"
AMPHIBIA = "NAME/NATURAL_OBJECT/LIVING_THING/AMPHIBIA"
BIRD = "NAME/NATURAL_OBJECT/LIVING_THING/BIRD"
FISH = "NAME/NATURAL_OBJECT/LIVING_THING/FISH"
FLORA = "NAME/NATURAL_OBJECT/LIVING_THING/FLORA"
FUNGUS = "NAME/NATURAL_OBJECT/LIVING_THING/FUNGUS"
INSECT = "NAME/NATURAL_OBJECT/LIVING_THING/INSECT"
LIVING_THING_OTHER = "NAME/NATURAL_OBJECT/LIVING_THING/LIVING_THING_OTHER"
MAMMAL = "NAME/NATURAL_OBJECT/LIVING_THING/MAMMAL"
MOLLUSC_ARTHROPOD = "NAME/NATURAL_OBJECT/LIVING_THING/MOLLUSC_ARTHROPOD"
REPTILE = "NAME/NATURAL_OBJECT/LIVING_THING/REPTILE"
LIVING_THING_PART = "NAME/NATURAL_OBJECT/LIVING_THING_PART"
ANIMAL_PART = "NAME/NATURAL_OBJECT/LIVING_THING_PART/ANIMAL_PART"
FLORA_PART = "NAME/NATURAL_OBJECT/LIVING_THING_PART/FLORA_PART"
LIVING_THING_PART_OTHER = (
    "NAME/NATURAL_OBJECT/LIVING_THING_PART/LIVING_THING_PART_OTHER"
)
MINERAL = "NAME/NATURAL_OBJECT/MINERAL"
NATURAL_OBJECT_OTHER = "NAME/NATURAL_OBJECT/NATURAL_OBJECT_OTHER"
ORGANIZATION = "NAME/ORGANIZATION"
CORPORATION = "NAME/ORGANIZATION/CORPORATION"
COMPANY = "NAME/ORGANIZATION/CORPORATION/COMPANY"
COMPANY_GROUP = "NAME/ORGANIZATION/CORPORATION/COMPANY_GROUP"
CORPORATION_OTHER = "NAME/ORGANIZATION/CORPORATION/CORPORATION_OTHER"
ETHNIC_GROUP = "NAME/ORGANIZATION/ETHNIC_GROUP"
ETHNIC_GROUP_OTHER = "NAME/ORGANIZATION/ETHNIC_GROUP/ETHNIC_GROUP_OTHER"
NATIONALITY = "NAME/ORGANIZATION/ETHNIC_GROUP/NATIONALITY"
FAMILY = "NAME/ORGANIZATION/FAMILY"
INTERNATIONAL_ORGANIZATION = "NAME/ORGANIZATION/INTERNATIONAL_ORGANIZATION"
ORGANIZATION_OTHER = "NAME/ORGANIZATION/ORGANIZATION_OTHER"
POLITICAL_ORGANIZATION = "NAME/ORGANIZATION/POLITICAL_ORGANIZATION"
CABINET = "NAME/ORGANIZATION/POLITICAL_ORGANIZATION/CABINET"
GOVERNMENT = "NAME/ORGANIZATION/POLITICAL_ORGANIZATION/GOVERNMENT"
MILITARY = "NAME/ORGANIZATION/POLITICAL_ORGANIZATION/MILITARY"
POLITICAL_ORGANIZATION_OTHER = (
    "NAME/ORGANIZATION/POLITICAL_ORGANIZATION/POLITICAL_ORGANIZATION_OTHER"
)
POLITICAL_PARTY = "NAME/ORGANIZATION/POLITICAL_ORGANIZATION/POLITICAL_PARTY"
SHOW_ORGANIZATION = "NAME/ORGANIZATION/SHOW_ORGANIZATION"
SPORTS_ORGANIZATION = "NAME/ORGANIZATION/SPORTS_ORGANIZATION"
PRO_SPORTS_ORGANIZATION = (
    "NAME/ORGANIZATION/SPORTS_ORGANIZATION/PRO_SPORTS_ORGANIZATION"
)
SPORTS_LEAGUE = "NAME/ORGANIZATION/SPORTS_ORGANIZATION/SPORTS_LEAGUE"
SPORTS_ORGANIZATION_OTHER = (
    "NAME/ORGANIZATION/SPORTS_ORGANIZATION/SPORTS_ORGANIZATION_OTHER"
)
PERSON = "NAME/PERSON"
PRODUCT = "NAME/PRODUCT"
ART = "NAME/PRODUCT/ART"
ART_OTHER = "NAME/PRODUCT/ART/ART_OTHER"
BOOK = "NAME/PRODUCT/ART/BOOK"
BROADCAST_PROGRAM = "NAME/PRODUCT/ART/BROADCAST_PROGRAM"
MOVIE = "NAME/PRODUCT/ART/MOVIE"
MUSIC = "NAME/PRODUCT/ART/MUSIC"
PICTURE = "NAME/PRODUCT/ART/PICTURE"
SHOW = "NAME/PRODUCT/ART/SHOW"
AWARD = "NAME/PRODUCT/AWARD"
CHARACTER = "NAME/PRODUCT/CHARACTER"
CLASS = "NAME/PRODUCT/CLASS"
CLOTHING = "NAME/PRODUCT/CLOTHING"
DECORATION = "NAME/PRODUCT/DECORATION"
DOCTRINE_METHOD = "NAME/PRODUCT/DOCTRINE_METHOD"
ACADEMIC = "NAME/PRODUCT/DOCTRINE_METHOD/ACADEMIC"
CULTURE = "NAME/PRODUCT/DOCTRINE_METHOD/CULTURE"
DOCTRINE_METHOD_OTHER = "NAME/PRODUCT/DOCTRINE_METHOD/DOCTRINE_METHOD_OTHER"
MOVEMENT = "NAME/PRODUCT/DOCTRINE_METHOD/MOVEMENT"
PLAN = "NAME/PRODUCT/DOCTRINE_METHOD/PLAN"
RELIGION = "NAME/PRODUCT/DOCTRINE_METHOD/RELIGION"
SPORT = "NAME/PRODUCT/DOCTRINE_METHOD/SPORT"
STYLE = "NAME/PRODUCT/DOCTRINE_METHOD/STYLE"
THEORY = "NAME/PRODUCT/DOCTRINE_METHOD/THEORY"
DRUG = "NAME/PRODUCT/DRUG"
FOOD = "NAME/PRODUCT/FOOD"
DISH = "NAME/PRODUCT/FOOD/DISH"
FOOD_OTHER = "NAME/PRODUCT/FOOD/FOOD_OTHER"
ID_NUMBER = "NAME/PRODUCT/ID_NUMBER"
LANGUAGE = "NAME/PRODUCT/LANGUAGE"
LANGUAGE_OTHER = "NAME/PRODUCT/LANGUAGE/LANGUAGE_OTHER"
NATIONAL_LANGUAGE = "NAME/PRODUCT/LANGUAGE/NATIONAL_LANGUAGE"
MATERIAL = "NAME/PRODUCT/MATERIAL"
MONEY_FORM = "NAME/PRODUCT/MONEY_FORM"
OFFENCE = "NAME/PRODUCT/OFFENCE"
PRINTING = "NAME/PRODUCT/PRINTING"
MAGAZINE = "NAME/PRODUCT/PRINTING/MAGAZINE"
NEWSPAPER = "NAME/PRODUCT/PRINTING/NEWSPAPER"
PRINTING_OTHER = "NAME/PRODUCT/PRINTING/PRINTING_OTHER"
PRODUCT_OTHER = "NAME/PRODUCT/PRODUCT_OTHER"
RULE = "NAME/PRODUCT/RULE"
LAW = "NAME/PRODUCT/RULE/LAW"
RULE_OTHER = "NAME/PRODUCT/RULE/RULE_OTHER"
TREATY = "NAME/PRODUCT/RULE/TREATY"
SERVICE = "NAME/PRODUCT/SERVICE"
STOCK = "NAME/PRODUCT/STOCK"
TITLE = "NAME/PRODUCT/TITLE"
POSITION_VOCATION = "NAME/PRODUCT/TITLE/POSITION_VOCATION"
TITLE_OTHER = "NAME/PRODUCT/TITLE/TITLE_OTHER"
UNIT = "NAME/PRODUCT/UNIT"
CURRENCY = "NAME/PRODUCT/UNIT/CURRENCY"
UNIT_OTHER = "NAME/PRODUCT/UNIT/UNIT_OTHER"
VEHICLE = "NAME/PRODUCT/VEHICLE"
AIRCRAFT = "NAME/PRODUCT/VEHICLE/AIRCRAFT"
CAR = "NAME/PRODUCT/VEHICLE/CAR"
SHIP = "NAME/PRODUCT/VEHICLE/SHIP"
SPACESHIP = "NAME/PRODUCT/VEHICLE/SPACESHIP"
TRAIN = "NAME/PRODUCT/VEHICLE/TRAIN"
VEHICLE_OTHER = "NAME/PRODUCT/VEHICLE/VEHICLE_OTHER"
WEAPON = "NAME/PRODUCT/WEAPON"
NUMEX = "NUMEX"
AGE = "NUMEX/AGE"
COUNTX = "NUMEX/COUNTX"
COUNTX_OTHER = "NUMEX/COUNTX/COUNTX_OTHER"
N_EVENT = "NUMEX/COUNTX/N_EVENT"
N_FACILITY = "NUMEX/COUNTX/N_FACILITY"
N_LOCATION = "NUMEX/COUNTX/N_LOCATION"
N_COUNTRY = "NUMEX/COUNTX/N_LOCATION/N_COUNTRY"
N_LOCATION_OTHER = "NUMEX/COUNTX/N_LOCATION/N_LOCATION_OTHER"
N_NATURAL_OBJECT = "NUMEX/COUNTX/N_NATURAL_OBJECT"
N_ANIMAL = "NUMEX/COUNTX/N_NATURAL_OBJECT/N_ANIMAL"
N_FLORA = "NUMEX/COUNTX/N_NATURAL_OBJECT/N_FLORA"
N_NATURAL_OBJECT_OTHER = "NUMEX/COUNTX/N_NATURAL_OBJECT/N_NATURAL_OBJECT_OTHER"
N_ORGANIZATION = "NUMEX/COUNTX/N_ORGANIZATION"
N_PERSON = "NUMEX/COUNTX/N_PERSON"
N_PRODUCT = "NUMEX/COUNTX/N_PRODUCT"
FREQUENCY = "NUMEX/FREQUENCY"
LATITUDE_LONGTITUDE = "NUMEX/LATITUDE_LONGTITUDE"
MEASUREMENT = "NUMEX/MEASUREMENT"
CALORIE = "NUMEX/MEASUREMENT/CALORIE"
INTENSITY = "NUMEX/MEASUREMENT/INTENSITY"
MEASUREMENT_OTHER = "NUMEX/MEASUREMENT/MEASUREMENT_OTHER"
PHYSICAL_EXTENT = "NUMEX/MEASUREMENT/PHYSICAL_EXTENT"
SEISMIC_INTENSITY = "NUMEX/MEASUREMENT/SEISMIC_INTENSITY"
SEISMIC_MAGNITUDE = "NUMEX/MEASUREMENT/SEISMIC_MAGNITUDE"
SPACE = "NUMEX/MEASUREMENT/SPACE"
SPEED = "NUMEX/MEASUREMENT/SPEED"
TEMPERATURE = "NUMEX/MEASUREMENT/TEMPERATURE"
VOLUME = "NUMEX/MEASUREMENT/VOLUME"
WEIGHT = "NUMEX/MEASUREMENT/WEIGHT"
MONEY = "NUMEX/MONEY"
MULTIPLICATION = "NUMEX/MULTIPLICATION"
NUMEX_OTHER = "NUMEX/NUMEX_OTHER"
ORDINAL_NUMBER = "NUMEX/ORDINAL_NUMBER"
PERCENT = "NUMEX/PERCENT"
POINT = "NUMEX/POINT"
RANK = "NUMEX/RANK"
SCHOOL_AGE = "NUMEX/SCHOOL_AGE"
STOCK_INDEX = "NUMEX/STOCK_INDEX"
TIME_TOP = "TIME_TOP"
PERIODX = "TIME_TOP/PERIODX"
PERIODX_OTHER = "TIME_TOP/PERIODX/PERIODX_OTHER"
PERIOD_DAY = "TIME_TOP/PERIODX/PERIOD_DAY"
PERIOD_MONTH = "TIME_TOP/PERIODX/PERIOD_MONTH"
PERIOD_TIME = "TIME_TOP/PERIODX/PERIOD_TIME"
PERIOD_WEEK = "TIME_TOP/PERIODX/PERIOD_WEEK"
PERIOD_YEAR = "TIME_TOP/PERIODX/PERIOD_YEAR"
TIMEX = "TIME_TOP/TIMEX"
DATE = "TIME_TOP/TIMEX/DATE"
DAY_OF_WEEK = "TIME_TOP/TIMEX/DAY_OF_WEEK"
ERA = "TIME_TOP/TIMEX/ERA"
TIME = "TIME_TOP/TIMEX/TIME"
TIMEX_OTHER = "TIME_TOP/TIMEX/TIMEX_OTHER"
TIME_TOP_OTHER = "TIME_TOP/TIME_TOP_OTHER"

ALL_LABELS = [
    NAME,
    COLOR,
    COLOR_OTHER,
    NATURE_COLOR,
    DISEASE,
    ANIMAL_DISEASE,
    DISEASE_OTHER,
    EVENT,
    EVENT_OTHER,
    INCIDENT,
    INCIDENT_OTHER,
    WAR,
    NATURAL_PHENOMENON,
    EARTHQUAKE,
    NATURAL_DISASTER,
    NATURAL_PHENOMENON_OTHER,
    OCCASION,
    CONFERENCE,
    GAME,
    OCCASION_OTHER,
    RELIGIOUS_FESTIVAL,
    FACILITY,
    ARCHAEOLOGICAL_PLACE,
    ARCHAEOLOGICAL_PLACE_OTHER,
    TUMULUS,
    FACILITY_OTHER,
    FACILITY_PART,
    GOE,
    AIRPORT,
    AMUSEMENT_PARK,
    CAR_STOP,
    GOE_OTHER,
    MARKET,
    MUSEUM,
    PARK,
    PORT,
    PUBLIC_INSTITUTION,
    RESEARCH_INSTITUTE,
    SCHOOL,
    SPORTS_FACILITY,
    STATION,
    THEATER,
    WORSHIP_PLACE,
    ZOO,
    LINE,
    BRIDGE,
    CANAL,
    LINE_OTHER,
    RAILROAD,
    ROAD,
    TUNNEL,
    WATER_ROUTE,
    GOD,
    LOCATION,
    ADDRESS,
    ADDRESS_OTHER,
    EMAIL,
    PHONE_NUMBER,
    POSTAL_ADDRESS,
    URL,
    ASTRAL_BODY,
    ASTRAL_BODY_OTHER,
    CONSTELLATION,
    PLANET,
    STAR,
    GEOLOGICAL_REGION,
    BAY,
    GEOLOGICAL_REGION_OTHER,
    ISLAND,
    LAKE,
    MOUNTAIN,
    RIVER,
    SEA,
    GPE,
    CITY,
    COUNTRY,
    COUNTY,
    GPE_OTHER,
    PROVINCE,
    LOCATION_OTHER,
    REGION,
    CONTINENTAL_REGION,
    DOMESTIC_REGION,
    REGION_OTHER,
    SPA,
    NAME_OTHER,
    NATURAL_OBJECT,
    COMPOUND,
    ELEMENT,
    LIVING_THING,
    AMPHIBIA,
    BIRD,
    FISH,
    FLORA,
    FUNGUS,
    INSECT,
    LIVING_THING_OTHER,
    MAMMAL,
    MOLLUSC_ARTHROPOD,
    REPTILE,
    LIVING_THING_PART,
    ANIMAL_PART,
    FLORA_PART,
    LIVING_THING_PART_OTHER,
    MINERAL,
    NATURAL_OBJECT_OTHER,
    ORGANIZATION,
    CORPORATION,
    COMPANY,
    COMPANY_GROUP,
    CORPORATION_OTHER,
    ETHNIC_GROUP,
    ETHNIC_GROUP_OTHER,
    NATIONALITY,
    FAMILY,
    INTERNATIONAL_ORGANIZATION,
    ORGANIZATION_OTHER,
    POLITICAL_ORGANIZATION,
    CABINET,
    GOVERNMENT,
    MILITARY,
    POLITICAL_ORGANIZATION_OTHER,
    POLITICAL_PARTY,
    SHOW_ORGANIZATION,
    SPORTS_ORGANIZATION,
    PRO_SPORTS_ORGANIZATION,
    SPORTS_LEAGUE,
    SPORTS_ORGANIZATION_OTHER,
    PERSON,
    PRODUCT,
    ART,
    ART_OTHER,
    BOOK,
    BROADCAST_PROGRAM,
    MOVIE,
    MUSIC,
    PICTURE,
    SHOW,
    AWARD,
    CHARACTER,
    CLASS,
    CLOTHING,
    DECORATION,
    DOCTRINE_METHOD,
    ACADEMIC,
    CULTURE,
    DOCTRINE_METHOD_OTHER,
    MOVEMENT,
    PLAN,
    RELIGION,
    SPORT,
    STYLE,
    THEORY,
    DRUG,
    FOOD,
    DISH,
    FOOD_OTHER,
    ID_NUMBER,
    LANGUAGE,
    LANGUAGE_OTHER,
    NATIONAL_LANGUAGE,
    MATERIAL,
    MONEY_FORM,
    OFFENCE,
    PRINTING,
    MAGAZINE,
    NEWSPAPER,
    PRINTING_OTHER,
    PRODUCT_OTHER,
    RULE,
    LAW,
    RULE_OTHER,
    TREATY,
    SERVICE,
    STOCK,
    TITLE,
    POSITION_VOCATION,
    TITLE_OTHER,
    UNIT,
    CURRENCY,
    UNIT_OTHER,
    VEHICLE,
    AIRCRAFT,
    CAR,
    SHIP,
    SPACESHIP,
    TRAIN,
    VEHICLE_OTHER,
    WEAPON,
    NUMEX,
    AGE,
    COUNTX,
    COUNTX_OTHER,
    N_EVENT,
    N_FACILITY,
    N_LOCATION,
    N_COUNTRY,
    N_LOCATION_OTHER,
    N_NATURAL_OBJECT,
    N_ANIMAL,
    N_FLORA,
    N_NATURAL_OBJECT_OTHER,
    N_ORGANIZATION,
    N_PERSON,
    N_PRODUCT,
    FREQUENCY,
    LATITUDE_LONGTITUDE,
    MEASUREMENT,
    CALORIE,
    INTENSITY,
    MEASUREMENT_OTHER,
    PHYSICAL_EXTENT,
    SEISMIC_INTENSITY,
    SEISMIC_MAGNITUDE,
    SPACE,
    SPEED,
    TEMPERATURE,
    VOLUME,
    WEIGHT,
    MONEY,
    MULTIPLICATION,
    NUMEX_OTHER,
    ORDINAL_NUMBER,
    PERCENT,
    POINT,
    RANK,
    SCHOOL_AGE,
    STOCK_INDEX,
    TIME_TOP,
    PERIODX,
    PERIODX_OTHER,
    PERIOD_DAY,
    PERIOD_MONTH,
    PERIOD_TIME,
    PERIOD_WEEK,
    PERIOD_YEAR,
    TIMEX,
    DATE,
    DAY_OF_WEEK,
    ERA,
    TIME,
    TIMEX_OTHER,
    TIME_TOP_OTHER,
]



import importlib
import sys

from wasabi import msg  # type: ignore

commands = ["train"]
MSG_AVAILABLE_COMMANDS = f"Available commands: {', '.join(commands)}"


def help(exits: int = 0):
    msg.info("usage: camphr <command>", MSG_AVAILABLE_COMMANDS, exits=exits)


def main():
    if len(sys.argv) == 1:
        help(1)
    cmd = sys.argv.pop(1)
    if cmd == "--help":
        help(0)
    if cmd not in commands:
        msg.fail(f"unknown command {cmd}.", MSG_AVAILABLE_COMMANDS, exits=1)
    else:
        run(cmd)


def run(cmd: str):
    # Lazy load `cmd` because Hydra's instantiation takes few seconds.
    # All `cmd` must exist in the form of `camphr.cli.cmd.main`.
    cmd = cmd.replace("-", "_")
    m = importlib.import_module(f"camphr.cli.{cmd}")
    getattr(m, "main")()


if __name__ == "__main__":
    main()



"""Udify: '75 Languages, 1 Model: Parsing Universal Dependencies Universally' (https://arxiv.org/abs/1904.02099)
"""
from typing import Dict, Iterable, Optional, Tuple

import spacy
import spacy.language
from spacy.language import Language
from spacy.pipeline.pipes import Sentencizer
from spacy.tokens import Doc

from camphr.types import Pathlike

from ..errors import Errors
from .allennlp_base import VALIDATION, AllennlpPipe
from .utils import flatten_docs_to_sents, set_heads


def load_udify_pipes(
    punct_chars: Optional[Iterable[str]] = None,
) -> Tuple[Sentencizer, "Udify"]:
    """Load udify pipe.

    Udify requires doc is sentenced, so `spacy.Sentencizer` is also created and returned.

    Returns:
        Pipes of (sentencizer, udify)

    Examples:
        >>> nlp = spacy.blank("ja")
        >>> pipes = load_udify_pipes(["。"])
        >>> [nlp.add_pipe(pipe) for pipe in pipes]
    """
    pipe = spacy.load("en_udify").get_pipe("udify")
    if punct_chars:
        sentencizer = Sentencizer(list(punct_chars))
    else:
        sentencizer = Sentencizer()
    return sentencizer, pipe


def load_udify(
    lang: str,
    punct_chars: Optional[Iterable[str]] = None,
    sentencizer: Optional[Sentencizer] = None,
) -> Language:
    """Load nlp with udify pipeline.

    Udify requires doc is sentenced, so `spacy.Sentencizer` is added to the `nlp`.

    Returns:
        `nlp` with udify and sentencizer.

    Examples:
        >>> nlp = load_udify("ja", punct_chars=["。"])
    """
    nlp = spacy.blank(lang)
    _sentencizer, udify = load_udify_pipes(punct_chars)
    nlp.add_pipe(sentencizer or _sentencizer)
    nlp.add_pipe(udify)
    return nlp


@spacy.component(
    "udify", assigns=["token.lemma", "token.dep", "token.pos", "token.head"]
)
class Udify(AllennlpPipe):
    @staticmethod
    def import_udify():
        try:
            from allennlp.common.util import import_submodules  # type: ignore
        except ImportError:
            Errors.E0("unofficial-allennlp-nightly")

        import_submodules("udify")

    @classmethod
    def from_archive(
        cls, archive_path: Pathlike, dataset_reader_to_load: str = VALIDATION
    ):
        cls.import_udify()
        return super().from_archive(archive_path, dataset_reader_to_load)

    def __init__(self, *args, **kwargs) -> None:
        self.import_udify()
        super().__init__(*args, **kwargs)

    def set_annotations(self, docs: Iterable[Doc], outputs: Dict):
        """Set udify's output, which is calculated in self.predict, to docs"""
        try:
            from udify.models.udify_model import OUTPUTS as UdifyOUTPUTS  # type: ignore
        except ImportError:
            Errors.E0("unofficial-udify")

        for sent, output in zip(flatten_docs_to_sents(docs), outputs):
            words = output[UdifyOUTPUTS.words]
            _doc_tokens = [token.text for token in sent]
            if words != _doc_tokens:
                raise ValueError(
                    "Internal error has occured."
                    f"Input text: {sent.text}\n"
                    f"Input tokens: {_doc_tokens}\n"
                    f"Model words: {words}"
                )

            for token, dep, upos, lemma in zip(
                sent,
                output[UdifyOUTPUTS.predicted_dependencies],
                output[UdifyOUTPUTS.upos],
                output[UdifyOUTPUTS.lemmas],
            ):
                token.dep_ = dep
                token.lemma_ = lemma
                token.pos_ = upos
            sent = set_heads(sent, output[UdifyOUTPUTS.predicted_heads])
            sent.doc.is_parsed = True



"""Base class and utils for AllenNLP"""
import pickle
import shutil
from pathlib import Path
from typing import Dict, Iterable

import torch
from spacy.pipeline import Pipe
from spacy.tokens import Doc

from camphr.pipelines.utils import flatten_docs_to_sents
from camphr.types import Pathlike

from ..errors import Errors

VALIDATION = "validation"
VALIDATION_DATASET_READER = "validation_dataset_reader"
DATASET_READER = "dataset_reader"
ARCHIVE = "archive"


class AllennlpPipe(Pipe):
    """Base class for allennlp Pipe.

    See .udify as an example.
    """

    def __init__(self, model=None, dataset_reader=None, archive_path=None, **cfg):
        self.model = model
        self.dataset_reader = dataset_reader
        self.cfg = cfg
        self.archive_path = archive_path

    @classmethod
    def from_archive(
        cls, archive_path: Pathlike, dataset_reader_to_load: str = VALIDATION
    ):
        """Construct from `allnlp.Archive`'s file."""
        # Uses lazy import because allennlp is an extra requirements.
        try:
            from allennlp.data import DatasetReader
            from allennlp.models.archival import load_archive
        except ImportError:
            Errors.E0("unofficial-allennlp-nightly")

        archive = load_archive(str(archive_path))
        config = archive.config
        if dataset_reader_to_load == VALIDATION and VALIDATION_DATASET_READER in config:
            dataset_reader_params = config[VALIDATION_DATASET_READER]
        else:
            dataset_reader_params = config[DATASET_READER]
        dataset_reader = DatasetReader.from_params(dataset_reader_params)
        return cls(
            model=archive.model,
            dataset_reader=dataset_reader,
            config={"allen_archive": archive.config},
            archive_path=Path(archive_path).absolute(),
        )

    def to_disk(self, path: Pathlike, **kwargs):
        path = Path(path)
        path.mkdir(exist_ok=True)
        if self.archive_path:
            shutil.copytree(self.archive_path, path / ARCHIVE)

        with (path / "cfg.pkl").open("wb") as f:
            pickle.dump(self.cfg, f)

    def from_disk(self, path: Pathlike, **kwargs) -> "AllennlpPipe":
        path = Path(path)
        archive_path = path / ARCHIVE
        new_self = self.from_archive(archive_path)
        with (path / "cfg.pkl").open("rb") as f:
            cfg = pickle.load(f)
        self.model = new_self.model
        self.dataset_reader = new_self.dataset_reader
        self.cfg.update(cfg)
        self.archive_path = archive_path
        return self

    def predict(self, docs: Iterable[Doc]) -> Dict:
        self.model.eval()
        all_sents = flatten_docs_to_sents(docs)
        with torch.no_grad():
            tokens_list = [[t.text for t in sent] for sent in all_sents]
            instances = [
                self.dataset_reader.text_to_instance(tokens) for tokens in tokens_list
            ]
            outputs = self.model.forward_on_instances(instances)
        return outputs



"""Defines utilities to get transformers attributes from name string."""
from typing import NamedTuple, Type

import transformers as trf


class _TrfMap(NamedTuple):
    name: str
    config: Type[trf.PretrainedConfig]
    tokenizer: Type[trf.PreTrainedTokenizer]
    model: Type[trf.PreTrainedModel]


_TRF_MAPS = (
    _TrfMap(
        "bert-base-japanese", trf.BertConfig, trf.BertJapaneseTokenizer, trf.BertModel
    ),
    _TrfMap(
        "xlm-roberta",
        trf.XLMRobertaConfig,
        trf.XLMRobertaTokenizer,
        trf.XLMRobertaModel,
    ),
    _TrfMap(
        "openai-gpt", trf.OpenAIGPTConfig, trf.OpenAIGPTTokenizer, trf.OpenAIGPTModel
    ),
    _TrfMap(
        "transfo-xl", trf.TransfoXLConfig, trf.TransfoXLTokenizer, trf.TransfoXLModel
    ),
    _TrfMap(
        "distilbert", trf.DistilBertConfig, trf.DistilBertTokenizer, trf.DistilBertModel
    ),
    _TrfMap(
        "camembert", trf.CamembertConfig, trf.CamembertTokenizer, trf.CamembertModel
    ),
    _TrfMap("albert", trf.AlbertConfig, trf.AlbertTokenizer, trf.AlbertModel),
    _TrfMap("roberta", trf.RobertaConfig, trf.RobertaTokenizer, trf.RobertaModel),
    _TrfMap("xlnet", trf.XLNetConfig, trf.XLNetTokenizer, trf.XLNetModel),
    _TrfMap("bert", trf.BertConfig, trf.BertTokenizer, trf.BertModel),
    _TrfMap("gpt2", trf.GPT2Config, trf.GPT2Tokenizer, trf.GPT2Model),
    _TrfMap("ctrl", trf.CTRLConfig, trf.CTRLTokenizer, trf.CTRLModel),
    _TrfMap("xlm", trf.XLMConfig, trf.XLMTokenizer, trf.XLMModel),
    _TrfMap("t5", trf.T5Config, trf.T5Tokenizer, trf.T5Model),
)

ALL_MODELS = [k.name for k in _TRF_MAPS]


def _get_trf_map(text: str) -> _TrfMap:
    for item in sorted(_TRF_MAPS, key=lambda x: len(x.name), reverse=True):
        if item.name in text:
            return item
    raise ValueError(f"Couldn't find any `_TrfMap` from {text}.")


def get_trf_name(text: str) -> str:
    return _get_trf_map(text).name


def get_trf_tokenizer_cls(text: str) -> Type[trf.PreTrainedTokenizer]:
    return _get_trf_map(text).tokenizer


def get_trf_config_cls(text: str) -> Type[trf.PretrainedConfig]:
    return _get_trf_map(text).config


def get_trf_model_cls(text: str) -> Type[trf.PreTrainedModel]:
    return _get_trf_map(text).model



from .maskedlm import (
    BertForMaskedLM,
    BertForMaskedLMPreprocessor,
    add_maskedlm_pipe,
    remove_maskedlm_pipe,
)
from .model import TrfModel
from .ner import TrfForNamedEntityRecognition
from .seq_classification import TrfForSequenceClassification
from .tokenizer import TrfTokenizer

__all__ = [
    "TrfModel",
    "TrfTokenizer",
    "BertForMaskedLMPreprocessor",
    "BertForMaskedLM",
    "TrfForNamedEntityRecognition",
    "TrfForSequenceClassification",
    "add_maskedlm_pipe",
    "remove_maskedlm_pipe",
]



"""Package camphr.pipelines defines spacy components."""
from .elmo import Elmo
from .pattern_search import PatternSearcher
from .regex_ruler import MultipleRegexRuler, RegexRuler
from .transformers import (
    TrfForNamedEntityRecognition,
    TrfForSequenceClassification,
    TrfModel,
    TrfTokenizer,
)
from .udify import Udify, load_udify

__all__ = [
    "Udify",
    "load_udify",
    "Elmo",
    "TrfModel",
    "TrfTokenizer",
    "TrfForSequenceClassification",
    "TrfForNamedEntityRecognition",
    "PatternSearcher",
    "MultipleRegexRuler",
    "RegexRuler",
]



"""Defines transformers tokenizer pipe."""
from typing import List, Optional, Sequence, Sized, cast

import spacy
import transformers
from spacy.pipeline import Pipe
from spacy.tokens import Doc
from spacy.vocab import Vocab

from .auto import get_trf_tokenizer_cls
from .utils import ATTRS, TransformersInput, TrfAutoMixin

TRANSFORMERS_TOKENIZER = "transformers_tokenizer"


@spacy.component(TRANSFORMERS_TOKENIZER)
class TrfTokenizer(TrfAutoMixin, Pipe):
    _TRF_NAME = "trf_name"
    _MODEL_CLS_GETTER = get_trf_tokenizer_cls

    def __init__(self, vocab: Vocab, model=True, **cfg):
        self.vocab = vocab
        self.model: transformers.PretrainedTokenizer = model
        self.cfg = cfg

    def predict(self, docs: Sequence[Doc]) -> TransformersInput:
        self.require_model()
        output = self.model.batch_encode_plus(
            [doc.text for doc in docs],
            add_special_tokens=True,
            return_tensors="pt",
            return_input_lengths=True,
            return_attention_masks=True,
            max_length=self.model.max_len,
            return_overflowing_tokens=True,
            return_special_tokens_mask=True,
        )
        return TransformersInput(**output)

    def set_annotations(self, docs: Sequence[Doc], inputs: TransformersInput):
        self.set_transformers_input(docs, inputs)
        self._set_tokens(docs, inputs)

    def update(self, docs: Sequence[Doc], *args, **kwargs):  # type: ignore
        y = self.predict(docs)
        self.set_annotations(docs, y)

    @staticmethod
    def set_transformers_input(docs: Sequence[Doc], inputs: TransformersInput):
        if docs:
            docs[0]._.set(ATTRS.batch_inputs, inputs)

    @staticmethod
    def get_transformers_input(docs: Sequence[Doc]) -> Optional[TransformersInput]:
        if docs:
            output: TransformersInput = docs[0]._.get(ATTRS.batch_inputs)
            if len(docs) != len(cast(Sized, output.input_ids)):
                ValueError(
                    f"""Mismatching batch size during processing.
                `docs` has {len(docs)} length, but `batch_inputs` has {len(output.input_ids)}.
                Maybe each component is processing with a different batch size.
                """
                )
            return output
        return None

    def _set_tokens(self, docs: Sequence[Doc], inputs: TransformersInput) -> None:
        for doc, x in zip(docs, inputs):
            tokens = self.model.convert_ids_to_tokens(x.input_ids)
            doc._.set(ATTRS.tokens, tokens)
            doc._.set(ATTRS.cleaned_tokens, self._clean_special_tokens(tokens))

    def _clean_special_tokens(self, tokens: List[str]) -> List[str]:
        return [
            token if token not in self.model.all_special_tokens else ""
            for token in tokens
        ]



"""Copied from http://svn.sourceforge.jp/svnroot/slothlib/CSharp/Version1/SlothLib/NLP/Filter/StopWord/word/Japanese.txt"""
STOP_WORDS = set(
    """
あそこ
あたり
あちら
あっち
あと
あな
あなた
あれ
いくつ
いつ
いま
いや
いろいろ
うち
おおまか
おまえ
おれ
がい
かく
かたち
かやの
から
がら
きた
くせ
ここ
こっち
こと
ごと
こちら
ごっちゃ
これ
これら
ごろ
さまざま
さらい
さん
しかた
しよう
すか
ずつ
すね
すべて
ぜんぶ
そう
そこ
そちら
そっち
そで
それ
それぞれ
それなり
たくさん
たち
たび
ため
だめ
ちゃ
ちゃん
てん
とおり
とき
どこ
どこか
ところ
どちら
どっか
どっち
どれ
なか
なかば
なに
など
なん
はじめ
はず
はるか
ひと
ひとつ
ふく
ぶり
べつ
へん
ぺん
ほう
ほか
まさ
まし
まとも
まま
みたい
みつ
みなさん
みんな
もと
もの
もん
やつ
よう
よそ
わけ
わたし
ハイ
上
中
下
字
年
月
日
時
分
秒
週
火
水
木
金
土
国
都
道
府
県
市
区
町
村


各
第
方
何
的
度
文
者
性
体
人
他
今
部
課
係
外
類
達
気
室
口
誰
用
界
会
首
男
女
別
話
私
屋
店
家
場
等
見
際
観
段
略
例
系
論
形
間
地
員
線
点
書
品
力
法
感
作
元
手
数
彼
彼女
子
内
楽
喜
怒
哀
輪
頃
化
境
俺
奴
高
校
婦
伸
紀
誌
レ
行
列
事
士
台
集
様
所
歴
器
名
情
連
毎
式
簿




回
匹
個
席
束
歳
目
通
面
円
玉
枚

前
後
左
右
次
先

春
夏
秋
冬



一
二
三
四
五
六
七
八
九
十
百
千
万
億
兆


下記
上記
時間
今回
前回
場合
一つ
年生
自分
ヶ所
ヵ所
カ所
箇所
ヶ月
ヵ月
カ月
箇月
名前
本当
確か
時点
全部
関係
近く
方法
我々
違い
多く
扱い
新た
その後
半ば
結局
様々
以前
以後
以降
未満
以上
以下
幾つ
毎日
自体
向こう
何人
手段
同じ
感じ
""".split()
)



"""Package ner_labels defines labels of named entity recognition and some utilities."""
from .labels_ene import ALL_LABELS as ene_labels
from .labels_irex import ALL_LABELS as irex_labels

LABELS = {"ene": ene_labels, "irex": irex_labels}

__all__ = ["LABELS"]



from spacy.language import Language
from spacy.pipeline.entityruler import EntityRuler

import camphr.ner_labels.labels_ontonotes as L

person_patterns = [
    {"label": L.PERSON, "pattern": [{"TAG": "名詞,固有名詞,人名,姓"}, {"TAG": "名詞,固有名詞,人名,名"}]},
    {"label": L.PERSON, "pattern": [{"TAG": "名詞,固有名詞,人名,姓"}]},
    {"label": L.PERSON, "pattern": [{"TAG": "名詞,固有名詞,人名,名"}]},
]

LABEL_PERSON = L.PERSON


def create_person_ruler(nlp: Language) -> EntityRuler:
    """Create entity ruler that extracts person name with regex.

    Notes:
        This component must be used with `camphr.lang.mecab.Japanese`.
        In order to imporove accuracy, it is recommended to create a Mecab user dictionaly. See scripts/mecab_person_dictionary.

    Examples:
        >>> nlp = camphr.lang.mecab.Japanes()
        >>> nlp.add_pipe(create_person_ruler(nlp))
        >>> doc = nlp("今日は田中と散歩に行った")
        >>> doc.ents
        (田中, )
    """
    ruler = EntityRuler(nlp)
    ruler.add_patterns(person_patterns)
    return ruler



"""Module utils defines utility for ner labels.

This module can be called directly to generete "labels_**.py" from yml file.

Examples:
    $ python utils.py
"""
import subprocess
from pathlib import Path
from typing import IO, Dict, Iterable, List, Optional, Union

import yaml
from typing_extensions import Literal

from camphr.types import Pathlike
from camphr.utils import get_labels

__dir__ = Path(__file__).parent


def make_ner_labels(
    entity_types: Iterable[str], type_: Literal["BIO", "BILUO"] = "BIO"
) -> List[str]:
    """Make BILUO or BIO style ner label from entity type list.

    Examples:
        >>> make_ner_labels(["PERSON"])
        ["-", "O", "U-PERSON", "B-PERSON", ...]
    """
    labels = ["-", "O"]
    prefixes = [p for p in type_ if p != "O"]
    for l in entity_types:
        for pref in prefixes:
            labels.append(pref + "-" + l)
    return list(dict.fromkeys(labels))  # unique while keep ordering


def get_ner_labels(labels: Union[List[str], Pathlike], type_="BIO"):
    labels = get_labels(labels)
    if all(l[:2] in {"-", "O", "I-", "B-", "L-", "U-"} for l in labels):
        return labels
    return make_ner_labels(labels, type_)


def yml_to_py(yml_path: Union[str, Path], py_path: Union[str, Path]):
    """Convert yml file that defines ner labels to python script."""
    yml_path = Path(yml_path)
    py_path = Path(py_path)
    with yml_path.open() as f, py_path.open("w") as fw:
        d = yaml.safe_load(f)
        keys = sorted(extract_keys(d))
        generate_py(fw, keys)


def extract_keys(d: Dict[str, Optional[Dict]], prefix: str = "") -> List[str]:
    """Nested key are described with slash: e.g. Animal.Bird -> Animal/Bird"""
    res = []
    for k, v in d.items():
        res.append((prefix + k).upper())
        if isinstance(v, dict):
            res += extract_keys(v, prefix + k + "/")
    return res


def _crean_slash(key: str) -> str:
    return key.split("/")[-1]


def generate_py(f: IO, keys: List[str]):
    for k in keys:
        f.write(f"{_crean_slash(k)} = '{k}'\n")
    f.write(f'\nALL_LABELS = [{" ,".join(map(_crean_slash, keys))}]\n')


PREFIX = "labels_"
if __name__ == "__main__":
    for y in __dir__.glob("*.yml"):
        p = __dir__ / (PREFIX + y.stem + ".py")
        yml_to_py(y, p)
    subprocess.run(["black", "."])



"""Defines pattern search pipeline based on ahocorasik."""
from typing import Dict, Generator, Iterable, Optional, Tuple, cast

import ahocorasick
import spacy
from spacy.tokens import Doc
from spacy.util import filter_spans
from typing_extensions import Literal

from camphr.utils import SerializationMixin, get_doc_char_span


@spacy.component("pattern_searcher")
class PatternSearcher(SerializationMixin):
    serialization_fields = [
        "model",
        "label_type",
        "custom_label",
        "custom_label_map",
        "destructive",
        "cfg",
    ]

    def __init__(
        self,
        model: Optional[ahocorasick.Automaton] = None,
        label_type: Optional[
            Literal["custom_label_map", "custom_label", "value", "matched"]
        ] = None,
        custom_label: Optional[str] = None,
        custom_label_map: Optional[Dict[str, str]] = None,
        destructive=False,
        **cfg
    ):
        self.model = model
        self.label_type = label_type
        self.custom_label = custom_label
        self.custom_label_map = custom_label_map
        self._validate_label()
        self.destructive = destructive
        self.cfg = cfg

        if custom_label:
            self.label_type = "custom_label"
            assert custom_label_map is None
            assert label_type == "custom_label" or label_type is None
        elif custom_label_map:
            self.label_type = "custom_label_map"
            assert custom_label is None
            assert label_type == "custom_label_map" or label_type is None
        if self.label_type is None:
            # default
            self.label_type = "matched"

    def _validate_label(self):
        if self.custom_label:
            self.label_type = "custom_label"
        elif self.custom_label_map:
            self.label_type = "custom_label_map"
        assert self.label_type != "custom_label_map" or self.custom_label_map
        assert self.label_type != "custom_label" or self.custom_label

    def get_label(self, item: str) -> str:
        if self.label_type == "matched":
            return "matched"
        if self.label_type == "value":
            return item
        if self.label_type == "custom_label":
            return cast(str, self.custom_label)
        if self.label_type == "custom_label_map":
            assert self.custom_label_map is not None
            return self.custom_label_map[item]

        raise ValueError("Internal Error")

    @classmethod
    def Model(cls, words: Iterable[str], **cfg) -> ahocorasick.Automaton:
        model = ahocorasick.Automaton()
        for word in words:
            model.add_word(word, word)
        model.make_automaton()
        return model

    @classmethod
    def from_words(cls, words: Iterable[str], **cfg) -> "PatternSearcher":
        model = cls.Model(words)
        return cls(model, **cfg)

    @classmethod
    def from_nlp(cls, *args, **kwargs) -> "PatternSearcher":
        return cls()

    def get_char_spans(self, text: str) -> Generator[Tuple[int, int, str], None, None]:
        self.require_model()
        for j, word in cast(ahocorasick.Automaton, self.model).iter(text):
            i = j - len(word) + 1
            yield i, j + 1, word

    def __call__(self, doc: Doc) -> Doc:
        matches = self.get_char_spans(doc.text)
        spans = []
        for i, j, text in matches:
            span = get_doc_char_span(
                doc, i, j, destructive=self.destructive, label=self.get_label(text)
            )
            if span:
                spans.append(span)
        [s.text for s in spans]  # TODO: resolve the evaluation bug and remove this line
        ents = filter_spans(doc.ents + tuple(spans))
        doc.ents = tuple(ents)
        return doc



import json
from typing import IO, Any, Dict, Tuple

import fire
import tqdm

from camphr.utils import split_keepsep


class K:
    ents = "entities"


def _split_gold_jsonl(reader: IO[str], writer: IO[str], sep: str, verbose=False):
    for line in tqdm.tqdm(reader, disable=not verbose):
        data: Tuple[str, Dict[str, Any]] = json.loads(line)
        assert len(data) == 2, line
        text, gold = data
        assert isinstance(text, str), text
        assert isinstance(gold, dict), gold

        texts = split_keepsep(text, sep)
        if len(texts) == 1:
            writer.write(line)
            continue

        gold_ents = gold[K.ents]
        offset = 0
        for t in texts:
            next_offset = offset + len(t)
            ents = [
                (start - offset, end - offset, label)
                for start, end, label in gold_ents
                if start >= offset and end <= next_offset
            ]
            new_line = [t, {K.ents: ents}]
            writer.write(json.dumps(new_line) + "\n")
            offset = next_offset


def split_gold(fname: str, output: str, sep: str):
    with open(fname) as f, open(output, "w") as fw:
        _split_gold_jsonl(f, fw, sep, verbose=True)


if __name__ == "__main__":
    fire.Fire(split_gold)
